import type IVue from "vue";
import requireVue from "./require-vue";
import domReady from "./dom-ready";
import loadScript from "./load-script";
import loadStyleLink from "./load-style-link";
import { initSharing, registerShareModule, loadWfm } from "./wfm";
import strToJsVar from "./str-to-js-var";

// window.Hfv = {};
window.$HFC_CDN_URL = window.$HFC_CDN_URL || "https://cdn.jsdelivr.net/npm";
window.$HFC_WFM_CONTAINERS = window.$HFC_WFM_CONTAINERS || {};

function setup(Vue: typeof IVue, hfcToVue: any) {
  let mainTemplate: HTMLTemplateElement;
  let mainComponent: IVue.Component;

  let components: IVue.Component[] = [];
  const templates =
    document.querySelectorAll<HTMLTemplateElement>("template[hfv]");

  if (!templates.length) {
    return;
  } else if (templates.length === 1) {
    mainTemplate = templates[0];
    mainComponent = templateToComponent(mainTemplate as HTMLTemplateElement);
  } else {
    templates.forEach((template) => {
      const component = templateToComponent(template as HTMLTemplateElement);
      if (template.hasAttribute("main")) {
        if (mainComponent) {
          console.warn("[hfv] detect more than one main component");
        }

        mainTemplate = template;
        mainComponent = component;
        return;
      }

      components.push(component);
      template.parentNode?.removeChild(template);
    });
  }

  if (!mainTemplate! || !mainComponent!) {
    console.warn("[hfv] main template not found");
    return;
  }

  let container;
  let containerSelector = mainTemplate.getAttribute("container");
  if (containerSelector) {
    container = document.querySelector(containerSelector);
  }

  if (!container) {
    container = document.createElement("div");
    mainTemplate.parentNode?.insertBefore(container, mainTemplate);
  }

  container.setAttribute("hfv", "");

  const topComponents = (mainComponent as any).components || {};
  delete (mainComponent as any).components;
  components.forEach((item) => {
    topComponents[item.name!] = item;
  });

  // buildin components
  topComponents["hook"] = Vue.defineComponent({
    emits: ["mounted", "unmounted"],
    setup({}, { slots, emit }) {
      Vue.onMounted(() => emit("mounted"));
      Vue.onUnmounted(() => emit("unmounted"));
      return () => slots.default && slots.default();
    },
  });

  const app = Vue.createApp(mainComponent);

  Object.keys(topComponents).forEach((name) => {
    app.component(name, topComponents[name]);
  });

  app.mount(container);
  mainTemplate.parentNode?.removeChild(mainTemplate);

  function buildAsyncVueHfc(name: string, version: string) {
    return Vue.defineAsyncComponent({
      loader() {
        const target = "@hyper.fun/" + name + "@" + version;

        let cdnUrl = window.$HFC_CDN_URL;
        const rewrite = (window as any)[`$HFC_CDN_REWRITE_${target}`];
        if (rewrite) cdnUrl = rewrite;
        const componentUrl = `${cdnUrl}/${target}/wfm/entry.js`;

        return loadScript(componentUrl)
          .then(() => {
            loadStyleLink(`${cdnUrl}/${target}/hfc.css`);
            return loadWfm("@hyper.fun/" + name);
          })
          .then((comp: any) => {
            return hfcToVue(comp.default, true);
          })
          .catch((err) => {
            console.error(err);
            console.warn(`[hfv] faild to load component: ${name} ${version} `);
            return Vue.defineComponent({});
          });
      },
    });
  }

  function templateToComponent(t: HTMLTemplateElement) {
    const templateHtml = t.innerHTML;
    let data: any;
    const dataJsonStr = t.getAttribute("data-json");
    if (dataJsonStr) {
      try {
        data = JSON.parse(decodeURI(dataJsonStr));
      } catch (error) {
        console.warn("[hfv] fail to pares data-json");
        console.error(error);
      }
    }

    if (!data) {
      const dataStr = t.getAttribute(":data");
      if (dataStr) data = strToJsVar(dataStr);
    }

    data = data || {};

    const methods: Record<string, any> = {};
    const methodStr = t.getAttribute(":methods");
    if (methodStr) {
      const methodArr = strToJsVar(methodStr);
      if (Array.isArray(methodArr)) {
        methodArr.forEach((method) => {
          if (window[method]) methods[method] = window[method];
        });
      }
    }

    let components: Record<string, any> = {};

    const hfcTagNameRe = /<(\w+-[\w-]+)\s+v([\w\.-]+)=/g;
    let _tr;
    while ((_tr = hfcTagNameRe.exec(templateHtml))) {
      const [_, name, version] = _tr;
      components[name] = buildAsyncVueHfc(name, version);
    }

    const componentStr = t.getAttribute("imports");
    if (componentStr) {
      componentStr
        .trim()
        .split(";")
        .forEach((item) => {
          let alias, name, version;
          const arr = item.trim().split(" ");
          if (arr.length === 2) {
            [name, version] = arr;
          } else if (arr.length === 3) {
            [alias, name, version] = arr;
          }

          if (!name || !version || version[0] !== "v") return;
          components[alias || name] = buildAsyncVueHfc(name, version.slice(1));
        });
    }

    return Vue.defineComponent({
      name: t.getAttribute("name") || "",
      data: () => data,
      methods,
      components,
      render: Vue.compile(templateHtml, {
        runtimeGlobalName: "$HFC_VUE",
      }),
    });
  }
}

Promise.all([
  requireVue(),
  import(/* webpackMode: "eager" */ "@hyper-function/hfc-to-vue/dist/to-vue"),
  domReady,
  initSharing(),
])
  .then(([Vue, hfcToVue]) => {
    registerShareModule("vue", window.$HFC_VUE.version, window.$HFC_VUE);
    setup(Vue, hfcToVue.default(Vue));
  })
  .catch((err: any) => {
    console.warn("[hfv] faild to init");
    console.error(err);
  });
